"""
A formulation in original variables of a wedding seating problem

Authors: Stuart Mitchell 2010
"""
from time import time

import coinor.pulp as pulp
import coinor.dippy as dippy
from random import shuffle, seed

#for repeatability
seed(0)

def unhappiness(guest_a, guest_b):
    """
    Return the unhappiness (0 is the best) of allocating
    two guests together in the same table
    """
    return abs(ord(guest_a) - ord(guest_b))

def solve_problem(guests, max_table_size, max_tables):
    #time this application
    start_time = time()

    seating_model = dippy.DipProblem(
        "Wedding Seating Model (DIP)", pulp.LpMinimize)

    #create the set of possible tables
    tables = range(max_tables)

    #create the set of possible seatings
    possible_seatings = [(g, t) for g in guests
                                for t in tables]

    #create a binary variable to model if a guest sits
    #at a particular table
    x = pulp.LpVariable.dicts('possible_seatings',
                              possible_seatings, 
                              lowBound = 0,
                              upBound = 1,
                              cat = pulp.LpInteger)

    #create variable to model table unhappiness
    unhappy = pulp.LpVariable.dicts('table_happiness', tables,
                                    lowBound = 0,
                                    upBound = None,
                                    cat = pulp.LpContinuous)

    #minimize the total unhappiness of the tables
    seating_model += sum([unhappy[table] for table in tables])

    #specify the maximum number of guests per table
    for table in tables:
        seating_model.relaxation[table] += \
            sum([x[(guest, table)] for guest in guests]) <= \
            max_table_size, "Maximum_table_size_%s"%table

    #A guest must seated at one and only one table
    for guest in guests:
        seating_model += \
            sum([x[(guest, table)] for table in tables]) == \
            1, "Must_seat_%s"%guest

    #create a set of variables to model the objective function
    possible_pairs = [(a, b) for a in guests for b in guests
                      if ord(a) < ord(b)]

    #create constraints for each possible pair
    for (a, b) in possible_pairs:
        for table in tables:
            seating_model.relaxation[table] += \
                unhappy[table] >= \
                unhappiness(a, b) * \
                    (x[(a, table)] + x[(b, table)] - 1)

    def table_branch(self, sol):
        tol = 1e-10
        for guest in guests:
            for table in sorted(tables):
                assign = sol[x[(guest, table)]]
                if (assign > tol) and (assign < 1 - tol):
                    down_branch_ub = [(x[(guest, t)], 0)
                                      for t in tables if t > table]
                    up_branch_ub   = [(x[(guest, t)], 0)
                                      for t in tables if t <= table]

                    return ([], down_branch_ub, [], up_branch_ub)

##    seating_model.branch_method = table_branch

    def initialise_cols(self):
        initialAlg = 1
##        print "Generating initial cols..."
        unallocated = [guest for guest in guests]
        shuffle(unallocated)
        bvs = []

        if initialAlg == 0:
            tabGuests = {}
            for table in tables:
                tabGuests[table] = []
                
            while len(unallocated) > 0:
                for table in tables:
                    if len(unallocated) > 0:
                        guest = unallocated.pop()
                        tabGuests[table].append(guest)

            for table in tables:
                var_values = []
                if tabGuests[table]:
                    for guest in tabGuests[table]:
                        var_values.append((x[(guest,table)], 1.0))                
                    if len(tabGuests[table]) > 1:
                        tabUnhappiness = max(unhappiness(guest_a, guest_b)
                                             for guest_a in tabGuests[table]
                                             for guest_b in tabGuests[table]
                                             if ord(guest_a) < ord(guest_b))
                    else:
                        tabUnhappiness = 0.0
                    var_values.append((unhappy[table], tabUnhappiness))
                    dv = dippy.DecompVar(var_values, 0.0, tabUnhappiness)
##                    print dv.var_values, dv.cost
                    bvs.append((table, dv))

        else:
            for table in tables:
                tabGuests = []
                var_values = []
                for count in range(max_table_size):
                    if len(unallocated) > 0:
                        guest = unallocated.pop()
                        tabGuests.append(guest)
                        var_values.append((x[(guest,table)], 1.0))
                if tabGuests:
                    if len(tabGuests) > 1:
                        tabUnhappiness = max(unhappiness(guest_a, guest_b)
                                             for guest_a in tabGuests
                                             for guest_b in tabGuests
                                             if ord(guest_a) < ord(guest_b))
                    else:
                        tabUnhappiness = 0.0
                    var_values.append((unhappy[table], tabUnhappiness))
                    dv = dippy.DecompVar(var_values, 0.0, tabUnhappiness)
##                    print dv.var_values, dv.cost
                    bvs.append((table, dv))

        return bvs

##    seating_model.init_vars = initialise_cols

    #seating_model.writeLP('wedding_main.lp')
    #for table in tables:
    #    seating_model.writeRelaxed(table, 'wedding_relax%s.lp' % table);
        
    dippy.Solve(seating_model, {
           'doPriceCut' : '1',
        })

    end_time = time() - start_time
    return end_time, seating_model.objective.value(), tables, x, unhappy

if __name__ == '__main__':
    end_time, obj, tables, x, unhappy = solve_problem(guests, max_table_size, max_tables)
    for table in tables:
        print table,
        for guest in guests:
            if x[(guest,table)].value() >= 0.99:
                print guest,
        print unhappy[table].value()
        print
    print 'Total solution time for dippy model %s seconds'%end_time
